"""
The script edit from finc_multi_small_exon.R in the original codes
Find the possible missed small exons, with the threshold: small_exon_size <= 100 and abs(delta_ratio) <= 0.5. 
Each realign region will be output as a dict object with the following essential key values:
    read_i(query_i), read_ith_intron(query_ith_intron), tx_i(sub_i), contain_tx_ith_exon, sum_exon_size, contain_exon_num, chrom, strand, 
    realign_start, realign_end, read_intron_start(left_query_end), read_intron_end(right_query_start), tx_leftexon_end(left_sub_end), 
    tx_rightexon_start(right_sub_start), small_exon_starts, small_exon_ends, margin_len, margin_len_mod, delta_ratio, delta_ratio_mod
Each realign region corresponds to one read intron cover one or multiple continuous small exons of one transcript isoforms.
For one read intron overlaps with multiple transcripts isoforms, we can only keep one of them for realignment using the minimum of 
    abs(delta_ratio_mod) (param: simplify = True). As delta_ratio_mod is calculated in the later step, we will use the
    simplify parameter in the fix_small_exon.py. It can reduce the time of realignment if there are many regions overlaps with multiple 
    transcript isoforms. We can also use all of the realign region for realignment (param: simplify = False). The realignment result
    will only return the alignment with best realign score.
"""


import numpy as np
import pandas as pd


EMPTY_SS = pd.Series([])


def condense_exon(exons):
    """
    The function condenses the input dataframe into a series.
    The exon_i, exon_start, exon_end will be condensed into list.
    """
    exons.reset_index(inplace=True)
    cond_exons = exons.iloc[0].drop(["exon_i", "exon_start", "exon_end"])
    cond_exons["exon_ids"] = list(exons["exon_i"])
    cond_exons["exon_starts"] = list(exons["exon_start"])
    cond_exons["exon_ends"] = list(exons["exon_end"])
    return cond_exons


def find_overlap_tx(read_block, bed_df, ignore_strand=True):
    """
    The function use read_block to find which transcripts in bed_df overlaps with the range of introns in the read.
    Return the overlapped transcript index.
    """
    bed_df = bed_df[bed_df["chrom"] == read_block.chrom]
    if not ignore_strand:
        bed_df = bed_df[bed_df["strand"] == read_block.strand]
    left_flag = bed_df["end"] > min(read_block.introns["intron_starts"])
    right_flag = bed_df["start"] < max(read_block.introns["intron_ends"])
    tx_i = bed_df[left_flag & right_flag].index
    return tx_i


def interval_overlap(i1, i2):
    """
    The function judge whether two intervals overlap with each other. i1 and i2 should be list or tuple with two elements.
    We require the first element of the interval is start and the second is end. start < end. start and end are 0-based
    Return a bool. True for overlap.
    """
    if i1[0] > i1[1] or i2[0] > i2[1]:
        raise Exception("Error: not standard interval!", i1, i2)
    # We find minimap2 might have reads with intron connect insertion and then connect intron.
    if i1[0] == i1[1] or i2[0] == i2[1]:
        flag = False
    else:
        flag = i1[1] > i2[0] and i1[0] < i2[1]
    return flag


def find_cover_exons(read_block, exons, flank_len=20, small_exon_size=100):

    global EMPTY_SS

    def merge_check(row, ith_intron, intron_start, intron_end):
        # We need to filter out the ol_exon with exons at the end and the flank exons not overlaps with read exons.
        # Check the continuity of ol_exon (as we require exon size ≤ small_exon_size).
        end_check = (not 0 in row.exon_ids) and (
            not row.exon_num-1 in row.exon_ids)
        if not end_check:
            return EMPTY_SS
        leftmost_exon_id = min(row.exon_ids)
        rightmost_exon_id = max(row.exon_ids)
        contain_exon_num = len(row.exon_ids)
        cont_check = rightmost_exon_id-leftmost_exon_id+1 == contain_exon_num
        if not cont_check:
            return EMPTY_SS
        exon_sizes = pd.Series(row.exon_ends) - pd.Series(row.exon_starts)
        size_check = all(exon_sizes <= small_exon_size)
        if not size_check:
            return EMPTY_SS
        left_tx_exon = exons.loc[(
            row.tx_i, leftmost_exon_id-1), ["exon_start", "exon_end"]]
        left_read_exon = [read_block.exons["exon_starts"]
                          [ith_intron], read_block.exons["exon_ends"][ith_intron]]
        left_check = interval_overlap(left_tx_exon, left_read_exon)
        right_tx_exon = exons.loc[(
            row.tx_i, rightmost_exon_id+1), ["exon_start", "exon_end"]]
        right_read_exon = [read_block.exons["exon_starts"]
                           [ith_intron+1], read_block.exons["exon_ends"][ith_intron+1]]
        right_check = interval_overlap(right_tx_exon, right_read_exon)
        overlap_check = left_check & right_check
        if not overlap_check:
            return EMPTY_SS
        realign_start = max(min(
            left_tx_exon[1], left_read_exon[1]) - flank_len, left_tx_exon[0], left_read_exon[0])
        realign_end = min(max(
            right_tx_exon[0], right_read_exon[0]) + flank_len, right_tx_exon[1], right_read_exon[1])
        realign_s = pd.Series({
            "read_i": read_block.read_i,
            "read_ith_intron": ith_intron,
            "tx_i": row.tx_i,
            "contain_tx_ith_exons": row.exon_ids,
            "exon_sizes": list(exon_sizes),
            "sum_exon_size": sum(exon_sizes),
            "contain_exon_num": contain_exon_num,
            "chrom": row.chrom,
            "read_strand": read_block.strand,
            "tx_strand": row.strand,
            "realign_start": realign_start,
            "realign_end": realign_end,
            "tx_leftexon_end": left_tx_exon[1],
            "tx_rightexon_start": right_tx_exon[0],
            "small_exon_starts": row.exon_starts,
            "small_exon_ends": row.exon_ends,
            "read_intron_start": intron_start,
            "read_intron_end": intron_end,
            "margin_len": intron_start - left_tx_exon[1] + right_tx_exon[0] - intron_end})
        realign_s["delta_ratio"] = (
            realign_s["margin_len"] - realign_s["sum_exon_size"]) / realign_s["sum_exon_size"]
        return realign_s
    #
    realign_list = []
    for ith_intron, intron_start, intron_end in zip(range(read_block.introns["intron_num"]),
                                                    read_block.introns["intron_starts"], read_block.introns["intron_ends"]):
        left_flag = exons["exon_start"] >= intron_start
        right_flag = exons["exon_end"] <= intron_end
        cover_exon = exons[left_flag & right_flag]
        if not cover_exon.empty:
            cond_cover_exon = cover_exon.groupby(
                level="tx_i").apply(condense_exon)
            for rt in cond_cover_exon.itertuples():
                realign_ss = merge_check(
                    rt, ith_intron, intron_start, intron_end)
                if not realign_ss.empty:
                    realign_list.append(realign_ss)
    if len(realign_list) == 0:
        return False
    else:
        return realign_list


def find_missed_exons(read_block, bed_df, exons, ignore_strand=True, flank_len=20, small_exon_size=100):
    """
    The function find transcript exons which overlap with the intron regions.
    Return a dataframe, each row is one intron with overlapped exons of one transcript.
    """
    if read_block.exons["exon_num"] == 1:  # Filter out singleten
        return False
    ol_tx_i = find_overlap_tx(read_block, bed_df, ignore_strand=ignore_strand)
    if ol_tx_i.empty:
        return False
    # Only used the exons on the transcript which overlaps with the introns to increase speed
    exons = exons.loc[(ol_tx_i, slice(None)), :]
    realign_list = find_cover_exons(
        read_block, exons, flank_len=flank_len, small_exon_size=small_exon_size)
    return realign_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import sys
import textwrap as _textwrap
from .fix_small_exon_multiproc import run_multiprocess
from .version import VERSION

class CustomHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ', '.join(action.option_strings) + ' ' + args_string

    def _fill_text(self, text, width, indent):
        return ''.join(indent + line for line in text.splitlines(keepends=True))

    def _split_lines(self, text, width):
        text = self._whitespace_matcher.sub(' ', text).strip()
        return _textwrap.wrap(text, 80)


"""
The script is used to import MisSER package. 
Run run_multiprocess function to find and fix small-exon-missed reads.
"""


def main():
    parser = argparse.ArgumentParser(
        description="Find and fix small-exon-missed region in noisy-long reads based on transcript annotation.",
        formatter_class=CustomHelpFormatter
    )
    group = parser.add_mutually_exclusive_group()
    # positional arguments
    parser.add_argument("inBam", help="Input original bam file.")
    parser.add_argument(
        "genomeFasta", 
        help="Reference genome fasta file, with fai index under the same directory."
        )
    parser.add_argument("annotBed", help="Annotated transcripts file in BED12 format.")
    parser.add_argument("outRegion", help="Output Region file, regions contain missed small exons.")
    # optional arguments
    parser.add_argument(
        "-v", "--version", help="Print version and exit.",
        action="version", version="MisSER {0}".format(VERSION)
    )
    parser.add_argument(
        "-c", "--coreNum", help="The number of cpu cores we used.",
        default=1, type=int, metavar="N"
    )
    parser.add_argument(
        "-s", "--exonSizeThd", 
        help="The threshold of exons size, ignore exons with size > exonSizeThd.",
        default=80, type=int, metavar="N"
    )
    parser.add_argument(
        "-d", "--deltaRatioThd",
        help="The threshold of absolute delta ratio, ignore abs(delta ratio) > deltaRatioThd.",
        default=0.5, type=float, metavar="float"
    )
    parser.add_argument(
        "-f", "--flankLen", help="The extended length on the both sides of realign region.",
        default=20, type=int, metavar="N"
    )
    parser.add_argument(
        "--strandSpecific", help="Only compare reads consistent with annotated strand.",
        action="store_true", default=False
    )
    parser.add_argument(
        "--allTranscripts", help="Return all possible missed exons among overlapping transcripts.",
        action="store_true", default=False
    )
    parser.add_argument(
        "--fixFlankLen", help="Flank length will not extend to cover the adjacent indels.",
        action="store_true", default=False
    )
    parser.add_argument(
        "--debugMode", help="Won't stop when meet an error in one read.",
        action="store_true", default=False
    )
    parser.add_argument(
        "--setTag", help="Set fr tags on the realigned reads",
        action="store_true", default=False
    )
    group.add_argument("-o", "--outBam", help="Output realigned bam file.", metavar="file")
    group.add_argument(
        "--onlyRegion", help="Only return the Region file without realign process.",
        action="store_true", default=False
    )
    # output help message when no param.
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    if not args.onlyRegion:
        if args.outBam is None:
            raise Exception(
                "Require -o or --outBam if --onlyRegion flag is not set!")
    run_multiprocess(
        args.inBam, args.genomeFasta, args.annotBed, args.outBam, args.outRegion,
        small_exon_size=args.exonSizeThd, flank_len=args.flankLen,
        ignore_strand=not(args.strandSpecific), nprocess=args.coreNum,
        delta_ratio_thd=args.deltaRatioThd, simplify=not(args.allTranscripts),
        float_flank_len=not(args.fixFlankLen), only_region=args.onlyRegion,
        debug_mode=args.debugMode, set_tag=args.setTag
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
BAM file will be input use pysam and each line in BAM file will be looped to find the possible missed small exons, with
    the threshold: small_exon_size <= 100 (default) and abs(delta_ratio) <= 0.5 (default).
The realign region will only be aligned to the exon sequences of the transcript, not compare with intron sequences.
"""

import csv
import re
import sys
import time
import warnings
import os
import logging

import numpy as np
import pandas as pd
import multiprocessing as mp
import pysam
import parasail
import collections

from itertools import groupby
from .block_class import (Cigar, Cigar_fromlist, ReadBlock, CigarSliceError)
from .find_small_exon import find_missed_exons


def groupby_unsorted(seq, key_fun=lambda x: x):
    """
    Unsorted version of groupby in itertools
    """
    indexs_dict = collections.defaultdict(list)
    for i, elem in enumerate(seq):
        indexs_dict[key_fun(elem)].append(i)
    for key, idxs in indexs_dict.items():
        yield key, [seq[i] for i in idxs]


def parse_bed(bed_path):
    """
    The function parse an annotation BED file to a dataframe contain list.
    We used to use list of dict as the result.
    [{'tx_i':146, 'chrom':'chrIS', 'strand':'+', 'start':6955486, 'end':6962816, 'name':'R1_101_1', 'block_count':3, 'block_sizes':[243, 116, 360], 'block_starts':[0, 4897, 6970]},
     {...},
     ...
     {...}]
    Actually, the list of dict is easy to be transformed to dataframe by pd.DataFrame(list_dict) and vice versa df.to_dict('records')
    The result dataframe should be sorted by start and end sites.
    """
    bed_col = [
        "chrom", "chromStart", "chromEnd", "name", "score", "strand",
        "thickStart", "thickEnd", "itemRgb", "blockCount", "blockSizes", "blockStarts"
    ]
    # Set the chrom as str because reference in ensembl do not have prefix 'chr'.
    # the read.reference_name in pysam is string.
    bed = pd.read_csv(
        bed_path, sep='\t', header=None, names=bed_col,
        quoting=csv.QUOTE_NONE,
        dtype={"chrom": str, "blockSizes": str, "blockStarts": str}
    )
    bed["tx_i"] = bed.index
    #
    bed_dict_list = bed.to_dict("records")
    parse_dict_list = []
    for tx_dict in bed_dict_list:
        tx_dict["blockSizes"] = [np.int64(x) for x in tx_dict["blockSizes"].split(',') if len(x) > 0]
        tx_dict["blockStarts"] = [np.int64(x) for x in tx_dict["blockStarts"].split(',')if len(x) > 0]
        tx_dict["blockNum"] = len(tx_dict["blockSizes"])
        parse_dict_list.append(tx_dict)
    bed = pd.DataFrame(parse_dict_list)
    bed_df = bed[["tx_i", "chrom", "strand"]]
    bed_df["tx"] = bed["name"]
    bed_df["start"] = bed["chromStart"]
    bed_df["end"] = bed["chromEnd"]
    bed_df["exon_sizes"] = bed["blockSizes"]
    bed_df["exon_starts"] = bed["blockStarts"]
    bed_df["exon_num"] = bed["blockNum"]
    return bed_df


def extend_tx(tx):
    """
    The function extends the list at exon_sizes and exon_starts in the transcript dataframe into exons.
    One line stands for one exon in the extended result.
    """
    if not all(tx["start"] < tx["end"]):  # We require start < end in the bed_df
        raise Exception("Error: start ≥ end in bed file!")
    exon_starts = tx.apply(
        lambda r: r["start"] + pd.Series(r["exon_starts"]), axis=1)
    exon_ends = tx.apply(
        lambda r: r["start"] + pd.Series(r["exon_starts"]) + pd.Series(r["exon_sizes"]), axis=1)
    exon_starts = exon_starts.stack().reset_index(level=1)
    exon_ends = exon_ends.stack().reset_index(level=1)
    if (not all(exon_starts["level_1"] == exon_ends["level_1"])) or (not all(exon_starts.index == exon_ends.index)):
        raise Exception("Error: extend exon fall!")
    exons = pd.DataFrame(
        {"exon_i": exon_starts["level_1"], "exon_start": exon_starts.iloc[:, 1], "exon_end": exon_ends.iloc[:, 1]}, dtype="int")
    exons = tx.drop(["exon_sizes", "exon_starts"], axis=1).join(exons)
    exons.set_index(["tx_i", "exon_i"], inplace=True)
    return exons


def trace_realign_region(row, read_cigar, read_start):
    """
    The function is used to trace the rdp and ci at the border of realign region and return a dict
    """
    realn_start_t = read_cigar.trace_ci_rdp(start_ci=(0, 0), start_pos=read_start,
                                            end_pos=row["realign_start"], start_rdp=0)
    if not realn_start_t[2]:
        raise Exception("Error: cannot trace to the realign border", row)
    realn_end_t = read_cigar.trace_ci_rdp(start_ci=realn_start_t[0], start_pos=row["realign_start"],
                                          end_pos=row["realign_end"], start_rdp=realn_start_t[1])
    if not realn_end_t[2]:
        raise Exception("Error: cannot trace to the realign border", row)
    return {"realign_start_ci": realn_start_t[0], "realign_end_ci": realn_end_t[0],
            "realign_start_rdp": realn_start_t[1], "realign_end_rdp": realn_end_t[1]}


def edit_margin_len(row, read_cigar):
    """
    The function is used to edit the margin length in the original realign_ss.
    Margin length is called margin_length in the old file.
    One insertion 'I' or 'D' will +1 or -1 in the margin length.
    """
    cigar_slice = read_cigar.cslice(
        start_ci=row["realign_start_ci"], end_ci=row["realign_end_ci"])
    new_margin_len = row["margin_len"]
    for n, c in cigar_slice:
        if c == 'D':
            new_margin_len -= n
        elif c == 'I':
            new_margin_len += n
    return new_margin_len


def extend_realign_region(row, read_cigar):
    """
    This part is separated from the original modify_realign_region.

    Because minimap use bonus strategy on the splice site, There might be a long deletion or insertion beside the
    splice site. In case to make sure that the realign start do not begin with a deletion or the realign end do not
    end with a insertion and deletion, we need to extend the boundary of the realign region when the case happen.

    Shown as follows:

    Deletion:
    Start:  ... End:
      # @         @ #
    ATGCA       GCTTC
    AT--A       GC--C
        @           @

    Insertion:
    Start:  ... End:
        @           @
    AT--A       GC--C
    ATGCA       GCTTC
      # @         @ #

    The symbol '@' means the original position of ref_start/end (above the sequence) and query_start/end (below the
    sequence); The symbol '#' means the position after modification.
    """
    try:
        left_slice_group = read_cigar.cslice(
            end_ci=row["realign_start_ci"])[-1]
        if left_slice_group[1] in {'D', 'I'}:
            if left_slice_group[1] == 'D':
                row["realign_start"] -= left_slice_group[0]
            else:
                row["realign_start_rdp"] -= left_slice_group[0]
            if row["realign_start_ci"][1] == 0:
                row["realign_start_ci"] = (row["realign_start_ci"][0]-1, 0)
            else:
                row["realign_start_ci"] = (row["realign_start_ci"][0], 0)
    except IndexError:  # most case because realign_start_ci is (0, 0)
        pass
    try:
        right_slice_group = read_cigar.cslice(
            start_ci=row["realign_end_ci"])[0]
        if right_slice_group[1] in {'D', 'I'}:
            if right_slice_group[1] == 'D':
                row["realign_end"] += right_slice_group[0]
            else:
                row["realign_end_rdp"] += right_slice_group[0]
            row["realign_end_ci"] = (row["realign_end_ci"][0]+1, 0)
    except IndexError:  # most case because realign_end_ci reach the cigar end.
        pass
    return row


def exclude_clip(row, read_cigar):
    """
    Separated from extend_realign_region, used to exclude the soft and hard clip at the start or end of the query_sequence.
    Because sometimes realign region might include the softclip region and the softclip region usually is really
    long which can lead to a wrong realignment result.
    Realignment will overwrite the hard clip, so we need to exclude them.
    """
    # Exclude the 'S' at both ends of the realign region
    cigar_slice = read_cigar.cslice(
        start_ci=row["realign_start_ci"], end_ci=row["realign_end_ci"])
    if cigar_slice[0][1] == 'S':
        row["realign_start_rdp"] += cigar_slice[0][0]
        row["realign_start_ci"] = (row["realign_start_ci"][0]+1, 0)
    if cigar_slice[0][1] == 'H':
        row["realign_start_ci"] = (row["realign_start_ci"][0]+1, 0)
    if cigar_slice[-1][1] == 'S':
        row["realign_end_rdp"] -= cigar_slice[0][0]
        row["realign_end_ci"] = (row["realign_end_ci"][0], 0)
    if cigar_slice[-1][1] == 'H':
        row["realign_end_ci"] = (row["realign_end_ci"][0], 0)
    return row


def modify_realign_region(row, read_cigar, read_start, float_flank_len):
    """
    The function is used to modify the realignment region.
    It also combines the trace, edit margin length and extend realign region process.
    Return: modified realign_ss
    """
    # Tracing process
    trace_dict = trace_realign_region(row[["realign_start", "realign_end", "read_intron_start",
                                           "read_intron_end"]], read_cigar, read_start)
    row = row.append(pd.Series(trace_dict))
    if float_flank_len:
        # Extend the boundary of realign region
        row = extend_realign_region(row, read_cigar)
    # Exclude soft clip
    row = exclude_clip(row, read_cigar)
    # Editing margin length
    row["margin_len_mod"] = edit_margin_len(row, read_cigar)
    row["delta_ratio_mod"] = (
        row["margin_len_mod"] - row["sum_exon_size"]) / row["sum_exon_size"]
    return row


def trace_realign_result(row, realign_cigar):
    """
    The function is used to trace the cigar index of introns which are used to split the realign_cigar.
    Return a ci_list.
    realn_tx_pos means the position regardless of introns and relative to the realign_start. The position of
    bases on the tx_seq.
    """
    ci_list = []
    intron_start = row["tx_leftexon_end"]
    realn_tx_pos = row["tx_leftexon_end"]-row["realign_start"]
    realn_intron_pos = [realn_tx_pos]
    for exon_size in row["exon_sizes"]:
        realn_tx_pos += exon_size
        realn_intron_pos.append(realn_tx_pos)
    for realn_tx_pos in realn_intron_pos:
        trace = realign_cigar.trace_ci_rdp(
            start_ci=(0, 0), start_pos=0, end_pos=realn_tx_pos, start_rdp=0)
        if not trace[2]:
            raise Exception(
                "Error:, cannot trace to the intron position in the realign cigar!")
        ci_list.append(trace[0])
    return ci_list


def cal_alignment_score(cigar_list, read_seq, ref_seq, cm, mm, go, ge):
    """
    mm, go and ge usually are negative to give penalty.
    """
    score = 0
    pos = 0
    rdp = 0
    for n, c in cigar_list:
        if n < 1:
            raise Exception(
                "Error: list in cigar is not standard!", cigar_list)
        if c == 'M':
            for _ in range(n):
                if read_seq[rdp] == ref_seq[pos]:
                    score += cm
                else:
                    score += mm
                pos += 1
                rdp += 1
        elif c == '=':
            score += cm*n
            pos += n
            rdp += n
        elif c == 'X':
            score += mm*n
            pos += n
            rdp += n
        elif c in {'I'}:
            score += go + ge*(n-1)
            rdp += n
        elif c == 'D':
            score += go + ge*(n-1)
            pos += n
        elif c == 'N':
            pos += n  # we do not give penalty on introns
        elif c == 'P':
            warnings.warn("Warning: 'P' in the cigar!", c)
        else:
            raise Exception(
                "Error: cigar character exceed assumption {'M', '=', 'X', 'I', 'D', 'P', 'N'}!", c)
    if rdp != len(read_seq) or pos != len(ref_seq):
        raise Exception("Error: input cigar not correspond to read_seq or ref_seq!", '\n',
                        cigar_list, '\n',
                        read_seq, '\n',
                        ref_seq)
    return score


def check_cigar_seq(cigar, read_seq, ref_seq):
    """
    The function is used to check whether the cigar is corresponded to read_seq and ref_seq
    """
    pos = 0
    rdp = 0
    for n, c in cigar.list:
        if c in {'M', 'X', '='}:
            pos += n
            rdp += n
        elif c in {'S', 'I'}:
            rdp += n
        elif c in {'D', 'N'}:
            pos += n
    if rdp != len(read_seq) or pos != len(ref_seq):
        raise Exception("Error: input cigar not correspond to read_seq or ref_seq!", '\n',
                        cigar.list, '\n',
                        read_seq, '\n',
                        ref_seq)
    return 0


def realign_read(read, read_cigar, realign_ss, ref_fa, score_matrix):
    """
    The function realign the region indicated in the realign_ss.
    Suffix rel means relative. pos_rel means the reference position relative to the realign start site.
    pos_rel = pos - realign_start.
    """
    check_cigar_seq(read_cigar, read.query_sequence, ref_fa.fetch(
        realign_ss["chrom"], start=read.reference_start, end=read.reference_end))
    realign_start = realign_ss["realign_start"]
    realign_end = realign_ss["realign_end"]
    # fetch half-open interval.
    # Use upper() to avoid soft-masked sequence with lowercase
    ref_seq = ref_fa.fetch(
        realign_ss["chrom"], start=realign_start, end=realign_end).upper()
    # tx_seq: part of ref_seq which does not contain intron sequences.
    # Also calculate transcript intron lengths
    intron_lens = []
    intron_start = realign_ss["tx_leftexon_end"]
    tx_seq = ref_seq[:realign_ss["tx_leftexon_end"]-realign_start]
    for s, e in zip(realign_ss["small_exon_starts"], realign_ss["small_exon_ends"]):
        tx_seq += ref_seq[s-realign_start:e-realign_start]
        intron_lens.append(s - intron_start)
        intron_start = e
    intron_lens.append(realign_ss["tx_rightexon_start"] - intron_start)
    tx_seq += ref_seq[realign_ss["tx_rightexon_start"]-realign_start:]
    read_seq = read.query_sequence[realign_ss["realign_start_rdp"]:realign_ss["realign_end_rdp"]]
    realign_result = parasail.nw_trace(read_seq, tx_seq, 1, 1, score_matrix)
    realign_score = realign_result.score
    realign_cigar_str = realign_result.cigar.decode.decode("utf-8")
    # We find the parasail might confuse the "X" with "=", so we convert all "X" and "=" to "M"
    realign_cigar_str = re.sub("[X=]", "M", realign_cigar_str)
    realign_cigar = Cigar(realign_cigar_str)
    realign_cigar.standardize()
    # check cal_alignment_score
    if realign_score != cal_alignment_score(realign_cigar.list, read_seq, tx_seq, 1, -1, -1, -1):
        print("read_seq:{0}\ntx_seq:{1}\nrealign_cigar_str:{2}\n".format(
            read_seq, tx_seq, realign_cigar_str
        ))
        raise Exception("Error: parasail alignment score different with the calculation!", realign_score,
                        cal_alignment_score(realign_cigar.list, read_seq, tx_seq, 1, -1, -1, -1))
    ori_realn_cigar_list = read_cigar.cslice(start_ci=realign_ss["realign_start_ci"],
                                             end_ci=realign_ss["realign_end_ci"])
    ori_score = cal_alignment_score(
        ori_realn_cigar_list, read_seq, ref_seq, 1, -1, -1, -1)
    increase_score = realign_score - ori_score
    realn_intron_cis = trace_realign_result(
        realign_ss[["tx_leftexon_end", "realign_start", "exon_sizes"]], realign_cigar)
    # joint new cigar
    new_cigar_list = read_cigar.cslice(end_ci=realign_ss["realign_start_ci"])
    start_ci = (0, 0)
    for ci, n in zip(realn_intron_cis, intron_lens):
        new_cigar_list.extend(realign_cigar.cslice(
            start_ci=start_ci, end_ci=ci))
        new_cigar_list.extend([[n, 'N']])
        start_ci = ci
    new_cigar_list.extend(realign_cigar.cslice(start_ci=start_ci))
    new_cigar_list.extend(read_cigar.cslice(
        start_ci=realign_ss["realign_end_ci"]))
    new_cigar = Cigar_fromlist(new_cigar_list)
    # check new cigar:
    # check_cigar_seq(new_cigar, read.query_sequence, ref_fa.fetch(realign_ss["chrom"], start=read.reference_start, end=read.reference_end))
    return new_cigar, increase_score


def find_fix(proc_region, ori_bam_path, out_bam_path, ref_fa_path, annot_bed, annot_exon,
             small_exon_size=80, flank_len=20, ignore_strand=True, set_tag=True,
             delta_ratio_thd=0.5, simplify=True, float_flank_len=False, only_region=False):
    """
    The function is used to find and fix small exon in one process.
    The proc_region is a three elements dict. {proc_id, start_read_id, end_read_id}
    Return a dict {"filter_list":[a list of realign_ss], "stat_dict":{a dict with statistics of misaligned reads}}
    """
    proc_id = proc_region["proc_id"]
    score_matrix = parasail.matrix_create("ACGT", 1, -1)
    ori_bam = pysam.AlignmentFile(ori_bam_path, "rb")
    if not os.path.exists("{0}.fai".format(ref_fa_path)):
        print("Build reference index...")
        pysam.faidx(ref_fa_path)
    ref_fa = pysam.FastaFile(ref_fa_path)
    # 0 based, left close right open
    start_read_id = proc_region["start_read_id"]
    end_read_id = proc_region["end_read_id"]
    if not only_region:
        out_bam_path_proc = out_bam_path + ".tmp{0}".format(proc_id)
        realign_bam = pysam.AlignmentFile(
            out_bam_path_proc, "wb", template=ori_bam)
    filter_list = []
    read_i = 0  # 0 based
    total_intron = 0
    misaligned_read_num = 0
    misaligned_intron_num = 0
    fix_read_num = 0
    fix_intron_num = 0
    for read in ori_bam.fetch(until_eof=True):
        if not(start_read_id <= read_i and read_i < end_read_id):
            read_i += 1
            continue
        if read.cigarstring is None or read.query_sequence is None:
            if not only_region:
                realign_bam.write(read)
            read_i += 1
            continue
        read_miss_exon_flag = False
        read_fix_flag = False
        read_block = ReadBlock(read, read_i)
        realign_list = find_missed_exons(read_block, annot_bed, annot_exon, ignore_strand=ignore_strand,
                                         flank_len=flank_len, small_exon_size=small_exon_size)
        if realign_list is not False:
            read_cigar = read_block.cigar
            read_start = read_block.starts[0]
            fix_region_num = 0  # count for fr tags
            for ith_intron, realign_intron_group in groupby_unsorted(realign_list, lambda x: x["read_ith_intron"]):
                realn_scores_increase = []
                realn_cigars = []
                delta_ratio_mods_abs = []
                intron_miss_exon_flag = False
                intron_fix_flag = False
                for i in range(len(realign_intron_group)):
                    realign_intron_group[i] = modify_realign_region(
                        realign_intron_group[i], read_cigar, read_start, float_flank_len)
                    delta_ratio_mods_abs.append(
                        abs(realign_intron_group[i]["delta_ratio_mod"]))
                if simplify:
                    min_ind = delta_ratio_mods_abs.index(
                        min(delta_ratio_mods_abs))
                    realign_intron_group = [realign_intron_group[min_ind]]
                for realign_ss in realign_intron_group:
                    # filter with delta_ratio
                    if abs(realign_ss["delta_ratio_mod"]) <= delta_ratio_thd:
                        read_miss_exon_flag = True
                        intron_miss_exon_flag = True
                        realign_ss["delta_ratio_flag"] = True
                        # add delta_ratio_flag when delta_ratio_mod <= delta_ratio_thd
                        if not only_region:
                            realign_result = realign_read(
                                read, read_cigar, realign_ss, ref_fa, score_matrix)
                            realign_ss["realn_increase_score"] = realign_result[1]
                            realign_ss["realn_flag"] = (True if realign_result[1] > 0 else False)
                            # add realn_flag when the region is fixed
                            realn_cigars.append(realign_result[0])
                            realn_scores_increase.append(realign_result[1])
                        filter_list.append(realign_ss)
                if intron_miss_exon_flag:
                    misaligned_intron_num += 1
                    if not only_region:
                        # when realigned region improve the alignment score, we will keep the realignment with max score.
                        if max(realn_scores_increase) > 0:
                            read_cigar = realn_cigars[realn_scores_increase.index(
                                max(realn_scores_increase))]
                            # check cigar string
                            # read_cigar.getstr()
                            read_fix_flag = True
                            intron_fix_flag = True
                            fix_intron_num += 1
                            fix_region_num += 1
            # if not only_region:
            #     new_cigar_str = read_cigar.getstr()
            #     read.cigarstring = new_cigar_str
        if not only_region:
            if read_fix_flag:
                new_cigar_str = read_cigar.getstr()
                read.cigarstring = new_cigar_str
                if set_tag:
                    if read.has_tag('fr'):
                        warnings.warn("Overwrite the existed fr tags on read:", read.qname)
                    read.set_tag('fr', fix_region_num, 'i')
            realign_bam.write(read)
        read_i += 1
        total_intron += read_block.introns["intron_num"]
        if read_miss_exon_flag:
            misaligned_read_num += 1
        if read_fix_flag:
            fix_read_num += 1
    stat_dict = {"read_count": read_i, "total_intron": total_intron,
                 "misaligned_read_num": misaligned_read_num,
                 "misaligned_intron_num": misaligned_intron_num,
                 "fix_read_num": fix_read_num, "fix_intron_num": fix_intron_num}
    # The read_count in stat_dict is the count of the whole bam file, not the part in the process.
    res_dict = {"filter_list": filter_list, "stat_dict": stat_dict}
    ori_bam.close()
    if not only_region:
        realign_bam.close()
        res_dict["out_bam_path_proc"] = out_bam_path_proc
    return res_dict


def find_fix_debug(proc_region, ori_bam_path, out_bam_path, ref_fa_path, annot_bed, annot_exon,
                   small_exon_size=80, flank_len=20, ignore_strand=True, set_tag=True,
                   delta_ratio_thd=0.5, simplify=True, float_flank_len=False, only_region=False):
    """
    The function is used to find and fix small exon in one process.
    The proc_region is a three elements dict. {proc_id, start_read_id, end_read_id}
    Return a dict {"filter_list":[a list of realign_ss], "stat_dict":{a dict with statistics of misaligned reads}}
    """
    proc_id = proc_region["proc_id"]
    score_matrix = parasail.matrix_create("ACGT", 1, -1)
    ori_bam = pysam.AlignmentFile(ori_bam_path, "rb")
    # define a bam to record the reads with error
    error_bam_path = ori_bam_path+".tmp{0}".format(proc_id)+"_err"
    error_bam = pysam.AlignmentFile(
        error_bam_path, "wb", template=ori_bam)
    error_case = 0
    # Check the exist of reference fasta index
    if not os.path.exists("{0}.fai".format(ref_fa_path)):
        print("Build reference index...")
        pysam.faidx(ref_fa_path)
    ref_fa = pysam.FastaFile(ref_fa_path)
    # 0 based, left close right open
    start_read_id = proc_region["start_read_id"]
    end_read_id = proc_region["end_read_id"]
    if not only_region:
        out_bam_path_proc = out_bam_path + ".tmp{0}".format(proc_id)
        realign_bam = pysam.AlignmentFile(
            out_bam_path_proc, "wb", template=ori_bam)
    filter_list = []
    read_i = 0  # 0 based
    total_intron = 0
    misaligned_read_num = 0
    misaligned_intron_num = 0
    fix_read_num = 0
    fix_intron_num = 0
    for read in ori_bam.fetch(until_eof=True):
        try:
            if not(start_read_id <= read_i and read_i < end_read_id):
                read_i += 1
                continue
            if read.cigarstring is None or read.query_sequence is None:
                if not only_region:
                    realign_bam.write(read)
                read_i += 1
                continue
            read_miss_exon_flag = False
            read_fix_flag = False
            read_block = ReadBlock(read, read_i)
            realign_list = find_missed_exons(
                read_block, annot_bed, annot_exon, ignore_strand=ignore_strand,
                flank_len=flank_len, small_exon_size=small_exon_size
            )
            if realign_list is not False:
                read_cigar = read_block.cigar
                read_start = read_block.starts[0]
                fix_region_num = 0
                for ith_intron, realign_intron_group in groupby_unsorted(realign_list, lambda x: x["read_ith_intron"]):
                    realn_scores_increase = []
                    realn_cigars = []
                    delta_ratio_mods_abs = []
                    intron_miss_exon_flag = False
                    intron_fix_flag = False
                    for i in range(len(realign_intron_group)):
                        realign_intron_group[i] = modify_realign_region(
                            realign_intron_group[i], read_cigar, read_start, float_flank_len)
                        delta_ratio_mods_abs.append(
                            abs(realign_intron_group[i]["delta_ratio_mod"]))
                    if simplify:
                        min_ind = delta_ratio_mods_abs.index(
                            min(delta_ratio_mods_abs))
                        realign_intron_group = [realign_intron_group[min_ind]]
                    for realign_ss in realign_intron_group:
                        # filter with delta_ratio
                        if abs(realign_ss["delta_ratio_mod"]) <= delta_ratio_thd:
                            read_miss_exon_flag = True
                            intron_miss_exon_flag = True
                            realign_ss["delta_ratio_flag"] = True
                            # add delta_ratio_flag when the region is found
                            if not only_region:
                                realign_result = realign_read(
                                    read, read_cigar, realign_ss, ref_fa, score_matrix)
                                realign_ss["realn_increase_score"] = realign_result[1]
                                realign_ss["realn_flag"] = (True if realign_result[1] > 0 else False)
                                # add realn_flag when the region is fixed
                                realn_cigars.append(realign_result[0])
                                realn_scores_increase.append(realign_result[1])
                            filter_list.append(realign_ss)
                    if intron_miss_exon_flag:
                        misaligned_intron_num += 1
                        if not only_region:
                            # when realigned region improve the alignment score, we will keep the realignment with max score.
                            if max(realn_scores_increase) > 0:
                                read_cigar = realn_cigars[realn_scores_increase.index(
                                    max(realn_scores_increase))]
                                # check cigar string
                                # read_cigar.getstr()
                                read_fix_flag = True
                                intron_fix_flag = True
                                fix_intron_num += 1
                                fix_region_num += 1
                # if not only_region:
                #     new_cigar_str = read_cigar.getstr()
                #     read.cigarstring = new_cigar_str
            if not only_region:
                if read_fix_flag:
                    new_cigar_str = read_cigar.getstr()
                    read.cigarstring = new_cigar_str
                    if set_tag:
                        if read.has_tag('fr'):
                            warnings.warn("Overwrite the existed fr tags on read:", read.qname)
                        read.set_tag('fr', fix_region_num, 'i')
                realign_bam.write(read)
            read_i += 1
            total_intron += read_block.introns["intron_num"]
            if read_miss_exon_flag:
                misaligned_read_num += 1
            if read_fix_flag:
                fix_read_num += 1
        except Exception as e:
            logging.exception(e)
            error_case += 1
            error_bam.write(read)
    stat_dict = {"read_count": read_i, "total_intron": total_intron,
                 "misaligned_read_num": misaligned_read_num,
                 "misaligned_intron_num": misaligned_intron_num,
                 "fix_read_num": fix_read_num, "fix_intron_num": fix_intron_num}
    # The read_count in stat_dict is the count of the whole bam file, not the part in the process.
    res_dict = {"filter_list": filter_list, "stat_dict": stat_dict}
    ori_bam.close()
    error_bam.close()
    if error_case == 0:
        os.remove(error_bam_path)
    if not only_region:
        realign_bam.close()
        res_dict["out_bam_path_proc"] = out_bam_path_proc
    return res_dict


def run_multiprocess(
    ori_bam_path, ref_fa_path, annot_bed_path, out_bam_path, out_realign_region_path,
    small_exon_size=80, flank_len=20, ignore_strand=True, set_tag=True, nprocess=1,
    delta_ratio_thd=0.5, simplify=True, float_flank_len=False, only_region=False, debug_mode=True
):
    start_time = time.process_time()
    start_time2 = time.time()
    annot_bed = parse_bed(annot_bed_path)
    annot_bed = annot_bed[annot_bed["exon_num"] != 1]  # Filter out singleten
    # In order to avoid repetitive computation, we extend the bed file in advance.
    annot_exon = extend_tx(annot_bed)
    read_count = int(pysam.view("-c", ori_bam_path).strip())
    filter_list = []
    if not only_region:
        out_bam_list = []
    total_intron = 0
    misaligned_read_num = 0
    misaligned_intron_num = 0
    fix_read_num = 0
    fix_intron_num = 0
    proc_list = []  # contain the list of proc_region
    per_read_num = int(read_count / nprocess)
    for i in range(nprocess):
        proc_region = {
            "proc_id": i,
            "start_read_id": i * per_read_num,
            "end_read_id": (i+1) * per_read_num
        }
        proc_list.append(proc_region)
    proc_list[-1]["end_read_id"] = read_count  # reset the end
    # def find_fix_proc(proc_region):
    #     # define a closure to use find_fix in multiprocessing.Pool instance
    #     res_dict = find_fix(
    #         proc_region, ori_bam_path, out_bam_path, ref_fa_path, annot_bed, score_matrix,
    #         small_exon_size=small_exon_size, flank_len=flank_len, ignore_strand=ignore_strand,
    #         delta_ratio_thd=delta_ratio_thd, simplify=simplify, float_flank_len=float_flank_len,
    #         only_region=only_region
    #     )
    #     return res_dict
    con_args = (
        ori_bam_path, out_bam_path, ref_fa_path, annot_bed, annot_exon,
        small_exon_size, flank_len, ignore_strand, set_tag,
        delta_ratio_thd, simplify, float_flank_len, only_region
    )
    pool = mp.Pool(processes=nprocess)
    if debug_mode:
        multi_res = [pool.apply_async(find_fix_debug, (proc_region, *con_args))
                     for proc_region in proc_list]
    else:
        multi_res = [pool.apply_async(find_fix, (proc_region, *con_args))
                     for proc_region in proc_list]
    pool.close()
    pool.join()
    res_list = [res.get() for res in multi_res]
    # res_list = pool.map(find_fix_proc, proc_list)
    for res_dict in res_list:
        filter_list.extend(res_dict["filter_list"])
        if not only_region:
            out_bam_list.append(res_dict["out_bam_path_proc"])
        total_intron += res_dict["stat_dict"]["total_intron"]
        misaligned_read_num += res_dict["stat_dict"]["misaligned_read_num"]
        misaligned_intron_num += res_dict["stat_dict"]["misaligned_intron_num"]
        fix_read_num += res_dict["stat_dict"]["fix_read_num"]
        fix_intron_num += res_dict["stat_dict"]["fix_intron_num"]
    if res_list[0]["stat_dict"]["read_count"] != read_count:
        print(
            "Warning: the total read count is not consist in pysam!",
            res_list[0]["stat_dict"]["read_count"],
            read_count
        )
    realign_df = pd.DataFrame(filter_list)
    realign_df.to_csv(out_realign_region_path, sep="\t",
                      index=False, quoting=csv.QUOTE_NONE)
    # Merge all output bam file in processes.
    pysam.merge("-cp", "-h", ori_bam_path, out_bam_path, *out_bam_list)
    for out_bam in out_bam_list:
        os.remove(out_bam)  # Remove tmp file
    end_time = time.process_time()
    end_time2 = time.time()
    print("Input BAM information:")
    print("\tTotal number of reads:", read_count)
    print("\tTotal number of introns:", total_intron)
    print("The result of finding misaligned exons:")
    if read_count == 0:
        print("\tThe number of misaligned reads:", misaligned_read_num)
    else:
        print("\tThe number of misaligned reads: {0}\tratio: {1:.2f}%".format(
            misaligned_read_num, misaligned_read_num/read_count*100
        ))
    if total_intron == 0:
        print("\tThe number of misaligned introns:", misaligned_intron_num)
    else:
        print("\tThe number of misaligned introns: {0}\tratio: {1:.2f}%".format(
            misaligned_intron_num, misaligned_intron_num/total_intron*100
        ))
    print("The result of realignment:")
    if misaligned_read_num == 0:
        print("\tThe number of fixed reads:", fix_read_num)
    else:
        print("\tThe number of fixed reads: {0}\tratio to misaligned reads: {1:.2f}%".format(
            fix_read_num, fix_read_num/misaligned_read_num*100
        ))
    if misaligned_intron_num == 0:
        print("\tThe number of fixed introns:", fix_intron_num)
    else:
        print("\tThe numberr of fixed introns: {0}\tratio to misaligned introns: {1:.2f}%".format(
            fix_intron_num, fix_intron_num/misaligned_intron_num*100
        ))
    print("Main process time used: {:.2f}s".format(end_time - start_time))
    print("Time used: {:.2f}s".format(end_time2 - start_time2))



import re
import warnings

import numpy as np

from copy import deepcopy


"""
The Cigar class is try to parse cigar string and achieve fast indexing.
    We define three kinds of positions of an aligned read:
    1. reference position: The position based on the reference. Normally, start or end without any suffix is reference position.
    2. read position (suffix: rdp or rdps): The position based on the read, which means the ith of the read base at this position.
    3. extend cigar position (suffix: ecp): The position based on the extend cigar.
    4. cigar index (suffix: ci): The index of cigar list.
    All of these positions are 0-based.
    Examples:
            * 
        012345678 (ecp)
        SSMMDDMMM
            *
          0123456 (pos)
        --ATTTGCC
              *
        0123  456 (rdp)
        GGAT--GGC

        If the alignment come the first star site (ecp=4, pos=2, rdp=4), The cigar character 'D' at ecp=4 will determine the
        alignment stage of 'T' at pos=2 and 'G' at ecp=4. As 'D' means deletion, so ecp and pos increase one and 
        point to the next one, while rdp remain unchanged rdp=4 .

               *
        012345678 (ecp)
        SSMMDDMMM
               *
          0123456 (pos)
        --ATTTGCC
               *
        0123  456 (rdp)
        GGAT--GGC

        If the alignment come the second star site (ecp=7, pos=5, rdp=5), The cigar character 'M' at ecp=7 will determine the
        alignment stage of 'C' at pos=5 and 'G' at ecp=5. As 'M' means match, so ecp, pos and rdp all increase one and 
        point to the next one.

    The relationship between pos, rdp, ecp can be writed as functions:
    pos = f1(ci)
    rdp = f2(ci)
    ci = g(pos, rdp)
"""


class CigarSliceError(Exception):
    """
    The class is used to raise an error when the cigar slice operation overstepped or start_ci > end_ci
    """

    def __init__(self, start_ci, end_ci, start_bound, end_bound):
        self.start_ci = start_ci
        self.end_ci = end_ci
        self.start_bound = start_bound
        self.end_bound = end_bound

    def __str__(self):
        return "cigar slice operation overstepped or start_ci > end_ci. \
                start_ci:{0}, start_bound:{1}, end_ci:{2}, end_bound:{3}" \
                    .format(self.start_ci, self.start_bound, self.end_ci, self.end_bound)

class CigarDefectError(Exception):
    def __init__(self, qname):
        self.qname = qname
    
    def __str__(self):
        return "Segement:{0} missing cigar field in the BAM file.".format(self.qname)


class Cigar(object):
    """
    We convert cigar string to cigar list. e.g. '6S55M77N' -> [[6, 'S'], [55, 'M'], [77, 'N']]
    In order to accelerate the tracing process. We abandon the method to use the extend cigar string and 
    the absolute extend cigar position (ecp). Instead, we index the cigar list at two levels (i, j).
    The first index i means the index of cigar list. e.g. i=1 trace to character group [55, 'M'] in the cigar list above
    The second index j means the index of  character. e.g. i=1 and j=10 trace to the 11th 'M' in [55, 'M']
    The cigar index (ci) is a tuple (i, j).
    """
    regexNum = re.compile('[0-9]+')
    regexChr = re.compile('[A-Z=]')
    cigarSet = {'M', 'I', 'D', 'N', 'S', 'H', 'P', '=', 'X'}

    def __init__(self, cstr):
        num_list = [int(x) for x in Cigar.regexNum.findall(cstr)]
        chr_list = Cigar.regexChr.findall(cstr)
        if not(set(chr_list) <= Cigar.cigarSet):
            raise Exception(
                "error: CIGAR contain the pattern exceed the assumption!", set(chr_list))
        self.str = cstr
        self.list = [[x, y] for x, y in zip(num_list, chr_list)]

    def __add_cumsum(self):
        self.cumsum_list = np.cumsum([x[0] for x in self.list])
    
    def getstr(self):
        self.str = ''
        for n, c in self.list:
            self.str = self.str + str(n) + c
        return self.str
    
    def ecp2ci(self, ecp):
        """
        Convert absolute extend cigar position to cigar index.
        Slow.
        """
        if not hasattr(self, 'cumsum_list'):
            self.__add_cumsum()
        if ecp < 0 or ecp > self.cumsum_list[-1]:
            raise Exception("Error: ecp out of bond!", ecp, self.cumsum_list[-1])
        for i, csum in zip(range(len(self.cumsum_list)), self.cumsum_list):
            if ecp < csum:
                break
        j = ecp - (self.cumsum_list[i-1] if i > 0 else 0)
        return (i, j)

    def ci2ecp(self, ci):
        """
        Convert cigar index to absolute extend cigar position.
        Quick.
        """
        if not hasattr(self, 'cumsum_list'):
            self.__add_cumsum()
        ecp = self.cumsum_list[ci[0] - 1] if ci[0] > 0 else 0
        ecp += ci[1]
        return ecp
    
    def standardize_list(cigar_list):
        """
        Check and standardize the cigar list. Include the following operation:
        1. merge adjacent cigar group with same kind of character.
        2. check and ensure the cigar number is positive. Delete 0 and raise error if negative.
        """
        if len(cigar_list) == 0:
            return []
        if cigar_list[0][0] == 0:
            return []
        standard_list = [deepcopy(cigar_list[0])]
        for n, c in cigar_list[1:]:
            if n == 0:
                continue
            elif n < 0:
                raise Exception("Error: negative cigar number!", cigar_list)
            if c == standard_list[-1][1]:
                standard_list[-1][0] += n
            else:
                standard_list.append([n, c])
        return standard_list

    def standardize(self):
        self.list = Cigar.standardize_list(self.list)
                

    def cslice(self, start_ci=None, end_ci=None):
        """
        start_ci: start cigar index
        end_ci: end cigar index
        Return a slice of cigar the list.
        Note: the cigar index is 0-based and the slice operation is half-open (i is closed but j is left-closed and right-open).
        An example:
        tc = Cigar("6S7M3D8M")
        tc.list:
        [[6, 'S'], [7, 'M'], [3, 'D'], [8, 'M']]
        tc.cslice(end_ci = (2, 0)) will return [[6, 'S'], [7, 'M']] 
        ATTENTION: The index (1, 7) is deprecated, we only allow the last cigar group can have index over-bound.
        """
        list_len = len(self.list)
        if start_ci is None:
            start_ci = (0, 0)
        if end_ci is None:
            end_ci = (list_len-1, self.list[-1][0])
        if start_ci == (list_len-1, self.list[-1][0]):
            return []
        # ci check
        if end_ci[0] == list_len-1:
            if not(0 <= start_ci[0] < list_len) or not(0 <= start_ci[1] < self.list[start_ci[0]][0]) \
            or not(0 <= end_ci[1] <= self.list[end_ci[0]][0]) or start_ci[0] > end_ci[0]:
                raise CigarSliceError(start_ci, end_ci, self.list[start_ci[0]][0], self.list[end_ci[0]][0])
        else:
            if not(0 <= start_ci[0] < list_len) or not(0 <= start_ci[1] < self.list[start_ci[0]][0]) \
            or not(end_ci[0] < list_len) or not(0 <= end_ci[1] < self.list[end_ci[0]][0]) or start_ci[0] > end_ci[0]:
                raise CigarSliceError(start_ci, end_ci, self.list[start_ci[0]][0], self.list[end_ci[0]][0])
        slice_list = deepcopy(self.list[start_ci[0]:end_ci[0]+1])
        if start_ci[0] == end_ci[0]:  # slice on the same one character group
            if end_ci[1] < start_ci[1]:
                raise CigarSliceError(start_ci, end_ci, self.list[start_ci[0]][0], self.list[end_ci[0]][0])
            slice_list[0][0] = end_ci[1] - start_ci[1]
        else:
            slice_list[0][0] = slice_list[0][0] - start_ci[1]
            if end_ci[1] == 0:
                slice_list = slice_list[:-1]
            else:
                slice_list[-1][0] = end_ci[1]
        slice_list = Cigar.standardize_list(slice_list)
        return slice_list

    def trace_ci_rdp(self, start_ci, start_pos, end_pos, start_rdp, end_rdp=None):
        """
        The defination of pos, rdp see The class ReadBlock.
        Actuallly, after the initial value start_ci, start_pos, start_rdp setted, we still need two arguments to 
        calculate end_ci (ci = g(pos, rdp)), end_pos and end_rdp. But usually, we only provide one argument, as 
        we only use end_pos to calculate end_ci, we will return end_ci and end_rdp as the pos first reach end_pos. 
        Return (end_ci, end_rdp, reach_flag)
        """
        pos = start_pos
        rdp = start_rdp
        ecp = self.ci2ecp(start_ci)
        reach_flag = False
        if start_pos == end_pos:
            reach_flag = True
            return (start_ci, start_rdp, reach_flag)
        if start_pos < end_pos:
            # forward tracing
            cigar_slice = self.cslice(start_ci=start_ci)
            for n, c in cigar_slice:
                if pos == end_pos:
                    reach_flag = True
                    break
                if c in {'M', 'X', '='}:
                    for j in range(n):
                        pos += 1
                        rdp += 1
                        ecp += 1
                        if pos == end_pos:
                            reach_flag = True
                            break
                elif c in {'S', 'I'}:
                    rdp += n
                    ecp += n
                elif c in {'D', 'N'}:
                    for j in range(n):
                        pos += 1
                        ecp += 1
                        if pos == end_pos:
                            reach_flag = True
                            break
                else:
                    ecp += n
        else:
            # backward tracing
            cigar_slice = self.cslice(end_ci=start_ci)
            for n, c in reversed(cigar_slice):
                if pos == end_pos:
                    reach_flag = True
                    break
                if c in {'M', 'X', '='}:
                    for j in range(n):
                        pos -= 1
                        rdp -= 1
                        ecp -= 1
                        if pos == end_pos:
                            reach_flag = True
                            break
                elif c in {'S', 'I'}:
                    for j in range(n):
                        rdp -= n
                        ecp -= n
                elif c in {'D', 'N'}:
                    for j in range(n):
                        pos -= 1
                        ecp -= 1
                        if pos == end_pos:
                            reach_flag = True
                            break
                else:
                    ecp -= n
        end_ci = self.ecp2ci(ecp)
        end_rdp = rdp
        return (end_ci, end_rdp, reach_flag)


class Cigar_fromlist(Cigar):
    """
    Create cigar object from cigar list rather than string.
    """

    def __init__(self, cigar_list):
        self.list = cigar_list

    

class ReadBlock(object):
    """
    The class is try to parse one read from BAM file (pysam.AlignedSegment) and return attribute of the blocks.
    To accelerate tracing process, we also return cigar list index to trace the introns.
    """

    def __init__(self, read, read_i):
        self.read_i = read_i
        self.chrom = read.reference_name
        self.strand = '-' if read.is_reverse else '+'
        self.cigar = Cigar(read.cigarstring)
        pos = read.reference_start
        rdp = 0
        self.starts = [pos]
        self.ends = []
        for n, c in self.cigar.list:
            if c in {'M', '=', 'X'}:
                pos += n
                rdp += n
            elif c in {'S', 'I'}:
                rdp += n
            elif c in {'D', 'N'}:
                if c == 'N':
                    self.ends.append(pos)
                    pos += n
                    self.starts.append(pos)
                else:
                    pos += n
        self.ends.append(pos)
        self.exons = {"exon_num": len(
            self.starts), "exon_starts": self.starts, "exon_ends": self.ends}
        self.introns = {"intron_num": len(self.starts)-1, "intron_starts": self.ends[:-1], "intron_ends": self.starts[1:]}



import setuptools
import codecs

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

exec(codecs.open("MisSER/version.py").read())


INSTALL_REQUIRES = [
    'numpy',
    'pandas',
    'pysam',
    'parasail>=1.1.17'
]

setuptools.setup(
    name='MisSER',
    version=VERSION,
    author='Zhen Liu',
    author_email='liuzhen2018@sibs.ac.cn',
    description='Find and fix missed small exons.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/zhenLiuExplr/MisSER-project',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    package_data={"MisSER": ["examples"]},
    package_dir={"MisSER": "MisSER"},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    python_requires='>=3',
    entry_points={
        'console_scripts': ['MisSER=MisSER.run_misser:main',],
    }
)
